# Plot.py -- Pamhyr
# Copyright (C) 2023-2024  INRAE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

import logging

from functools import reduce
from datetime import datetime
from numpy import sqrt

from tools import timer
from View.Tools.PamhyrPlot import PamhyrPlot

from View.Results.CustomPlot.Translate import CustomPlotTranslate

logger = logging.getLogger()

unit = {
    "elevation": "0-meter",
    "water_elevation": "0-meter",
    "discharge": "1-m3s",
    "velocity": "2-ms",
    "max_depth": "3-meter",
    "mean_depth": "3-meter",
    "froude": "4-dimensionless",
    "wet_area": "5-m2",
}


class CustomPlot(PamhyrPlot):
    def __init__(self, x, y, reach, profile, timestamp,
                 data=None, canvas=None, trad=None,
                 toolbar=None, parent=None):
        super(CustomPlot, self).__init__(
            canvas=canvas,
            trad=CustomPlotTranslate(),
            data=data,
            toolbar=toolbar,
            parent=parent
        )

        self._x = x
        self._y = y
        self._reach = reach
        self._profile = profile
        self._timestamp = timestamp

        logger.debug(
            "Create custom plot for: " +
            f"{x} -> {','.join(y)}: " +
            f"reach={reach}, profile={profile}, " +
            f"timestamp={timestamp}"
        )

        self._y_axes = sorted(
            set(
                map(
                    lambda y: unit[y],
                    self._y
                )
            )
        )

        self._axes = {}

    def _draw_rk(self):
        results = self.data
        reach = results.river.reach(self._reach)
        rk = reach.geometry.get_rk()
        z_min = reach.geometry.get_z_min()
        q = list(
            map(
                lambda p: p.get_ts_key(self._timestamp, "Q"),
                reach.profiles
            )
        )
        z = list(
            map(
                lambda p: p.get_ts_key(self._timestamp, "Z"),
                reach.profiles
            )
        )

        shift = 0
        compt = 0
        for ax in sorted(self._axes):
            if compt == 0:
                self._axes[ax].spines['left'].set_position(('outward', shift))
                compt += 1
            else:
                self._axes[ax].spines['right'].set_position(('outward', shift))
                shift += 60

        lines = {}
        if "elevation" in self._y:

            ax = self._axes[unit["elevation"]]
            line = ax.plot(
                rk, z_min,
                color='grey', lw=1.,
            )
            lines["elevation"] = line

        if "water_elevation" in self._y:

            ax = self._axes[unit["water_elevation"]]
            line = ax.plot(
                rk, z, lw=1.,
                color='blue',
            )
            lines["water_elevation"] = line

            if "elevation" in self._y:
                ax.fill_between(
                    rk, z_min, z,
                    color='blue', alpha=0.5, interpolate=True
                )

        if "discharge" in self._y:

            ax = self._axes[unit["discharge"]]
            line = ax.plot(
                rk, q, lw=1.,
                color='r',
            )
            lines["discharge"] = line

        if "velocity" in self._y:

            ax = self._axes[unit["velocity"]]
            v = list(
                map(
                    lambda p: p.geometry.speed(
                        p.get_ts_key(self._timestamp, "Q"),
                        p.get_ts_key(self._timestamp, "Z")),
                    reach.profiles
                )
            )

            line = ax.plot(
                rk, v, lw=1.,
                color='g',
            )
            lines["velocity"] = line

        if "max_depth" in self._y:

            ax = self._axes[unit["max_depth"]]
            d = list(
                map(
                    lambda p: p.geometry.max_water_depth(
                        p.get_ts_key(self._timestamp, "Z")),
                    reach.profiles
                )
            )
            line = ax.plot(
                rk, d,
                color='brown', lw=1.,
            )
            lines["max_depth"] = line

        if "mean_depth" in self._y:

            ax = self._axes[unit["mean_depth"]]
            d = list(
                map(
                    lambda p: p.geometry.mean_water_depth(
                        p.get_ts_key(self._timestamp, "Z")),
                    reach.profiles
                )
            )

            line = ax.plot(
                rk, d,
                color='orange', lw=1.,
            )
            lines["mean_depth"] = line

        if "froude" in self._y:

            ax = self._axes[unit["froude"]]
            fr = list(
                map(
                    lambda p:
                        p.geometry.speed(
                            p.get_ts_key(self._timestamp, "Q"),
                            p.get_ts_key(self._timestamp, "Z")) /
                        sqrt(9.81 * (
                            p.geometry.wet_area(
                                p.get_ts_key(self._timestamp, "Z")) /
                            p.geometry.wet_width(
                                p.get_ts_key(self._timestamp, "Z"))
                            )),
                    reach.profiles
                )
            )

            line = ax.plot(
                rk, fr, color='black', linestyle='--', lw=1.,
            )
            lines["froude"] = line

        if "wet_area" in self._y:

            ax = self._axes[unit["wet_area"]]
            d = list(
                map(
                    lambda p: p.geometry.wet_area(
                        p.get_ts_key(self._timestamp, "Z")),
                    reach.profiles
                )
            )

            line = ax.plot(
                rk, d,
                color='blue', linestyle='--', lw=1.,
            )
            lines["wet_area"] = line

        # Legend
        lns = reduce(
            lambda acc, line: acc + line,
            map(lambda line: lines[line], lines),
            []
        )
        labs = list(map(lambda line: self._trad[line], lines))
        self.canvas.axes.legend(lns, labs, loc="best")

    def _customize_x_axes_time(self, ts, mode="time"):
        # Custom time display
        nb = len(ts)
        mod = int(nb / 5)
        mod = mod if mod > 0 else nb

        fx = list(
            map(
                lambda x: x[1],
                filter(
                    lambda x: x[0] % mod == 0,
                    enumerate(ts)
                )
            )
        )

        if mode == "time":
            t0 = datetime.fromtimestamp(0)
            xt = list(
                map(
                    lambda v: (
                        str(
                            datetime.fromtimestamp(v) - t0
                        ).split(",")[0]
                        .replace("days", self._trad["days"])
                        .replace("day", self._trad["day"])
                    ),
                    fx
                )
            )
        else:
            xt = list(
                map(
                    lambda v: str(datetime.fromtimestamp(v).date()),
                    fx
                )
            )

        self.canvas.axes.set_xticks(ticks=fx, labels=xt, rotation=45)

    def _draw_time(self):
        results = self.data
        reach = results.river.reach(self._reach)
        profile = reach.profile(self._profile)

        shift = 0
        compt = 0
        for ax in sorted(self._axes):
            if compt == 0:
                self._axes[ax].spines['left'].set_position(('outward', shift))
                compt += 1
            else:
                self._axes[ax].spines['right'].set_position(('outward', shift))
                shift += 60

        ts = list(results.get("timestamps"))
        ts.sort()

        q = profile.get_key("Q")
        z = profile.get_key("Z")
        z_min = profile.geometry.z_min()
        ts_z_min = list(
            map(
                lambda ts:  z_min,
                ts
            )
        )

        lines = {}
        if "elevation" in self._y:
            # Z min is constant in time

            ax = self._axes[unit["elevation"]]

            line = ax.plot(
                ts, ts_z_min,
                color='grey', lw=1.
            )
            lines["elevation"] = line

        if "water_elevation" in self._y:

            ax = self._axes[unit["water_elevation"]]
            line = ax.plot(
                ts, z, lw=1.,
                color='b',
            )
            lines["water_elevation"] = line

            if "elevation" in self._y:

                ax.fill_between(
                    ts, ts_z_min, z,
                    color='blue', alpha=0.5, interpolate=True
                )

        if "discharge" in self._y:

            ax = self._axes[unit["discharge"]]
            line = ax.plot(
                ts, q, lw=1.,
                color='r',
            )
            lines["discharge"] = line

        if "velocity" in self._y:

            ax = self._axes[unit["velocity"]]
            v = list(
                map(
                    lambda q, z: profile.geometry.speed(q, z),
                    q, z
                )
            )

            line = ax.plot(
                ts, v, lw=1.,
                color='g',
            )
            lines["velocity"] = line

        if "max_depth" in self._y:

            ax = self._axes[unit["max_depth"]]
            d = list(
                map(lambda z: profile.geometry.max_water_depth(z), z)
            )

            line = ax.plot(
                ts, d,
                color='brown', lw=1.,
            )
            lines["max_depth"] = line

        if "mean_depth" in self._y:

            ax = self._axes[unit["mean_depth"]]
            d = list(
                map(lambda z: profile.geometry.mean_water_depth(z), z)
            )

            line = ax.plot(
                ts, d,
                color='orange', lw=1.,
            )
            lines["mean_depth"] = line

        if "froude" in self._y:

            ax = self._axes[unit["froude"]]
            d = list(
                map(lambda z, q:
                    profile.geometry.speed(q, z) /
                    sqrt(9.81 * (
                        profile.geometry.wet_area(z) /
                        profile.geometry.wet_width(z))
                    ), z, q)
                )

            line = ax.plot(
                ts, d, color='black', linestyle='--', lw=1.,
            )
            lines["froude"] = line

        if "wet_area" in self._y:

            ax = self._axes[unit["wet_area"]]
            d = list(
                map(lambda z: profile.geometry.wet_area(z), z)
            )

            line = ax.plot(
                ts, d, color='blue', linestyle='--', lw=1.,
            )
            lines["wet_area"] = line

        self._customize_x_axes_time(ts)

        # Legend
        lns = reduce(
            lambda acc, line: acc + line,
            map(lambda line: lines[line], lines),
            []
        )
        labs = list(map(lambda line: self._trad[line], lines))
        self.canvas.axes.legend(lns, labs, loc="best")

    @timer
    def draw(self):
        self.canvas.axes.cla()
        self.canvas.axes.grid(color='grey', linestyle='--', linewidth=0.5)

        if self.data is None:
            return

        self.canvas.axes.set_xlabel(
            self._trad[self._x],
            color='black', fontsize=10
        )

        self.canvas.axes.set_ylabel(
            self._trad[self._y_axes[0]],
            color='black', fontsize=10
        )

        self._axes[self._y_axes[0]] = self.canvas.axes
        for axes in self._y_axes[1:]:
            if axes in self._axes:
                self._axes[axes].clear()
                continue

            ax_new = self.canvas.axes.twinx()
            ax_new.set_ylabel(
                self._trad[axes],
                color='black', fontsize=10
            )
            self._axes[axes] = ax_new

        if self._x == "rk":
            self._draw_rk()
        elif self._x == "time":
            self._draw_time()
        if self._x == "rk":
            reach = self.data.river.reach(self._reach)
            profile = reach.profile(self._profile)
            x = profile.rk
        elif self._x == "time":
            x = self._timestamp

        self._current, = self.canvas.axes.plot(
            [x, x],
            self.canvas.axes.get_ylim(),
            # label=self.label_timestamp,
            color='grey',
            linestyle="dashed",
            lw=1.,
        )

        self.canvas.figure.canvas.draw_idle()
        if self.toolbar is not None:
            self.toolbar.update()

    @timer
    def update(self):
        if not self._init:
            self.draw()
            self.draw_current()
            return

    def set_reach(self, reach_id):
        self._reach = reach_id
        self._profile = 0

        self.update()

    def set_profile(self, profile_id):
        self._profile = profile_id

        if self._x != "rk":
            self.update()
        else:
            self.draw_current()

    def set_timestamp(self, timestamp):
        self._timestamp = timestamp

        if self._x != "time":
            self.update()
        else:
            self.draw_current()

    def draw_current(self):
        if self._x == "rk":
            reach = self.data.river.reach(self._reach)
            profile = reach.profile(self._profile)
            x = profile.rk
        elif self._x == "time":
            x = self._timestamp
        self._current.set_data([x, x], self.canvas.axes.get_ylim())
        self.canvas.figure.canvas.draw_idle()
